==> code2logic-1.0.2-py3-none-any.whl: code2logic/logicml.py <==
"""
LogicML Format Generator for Code2Logic.

Optimal hybrid format combining:
- YAML precision (signatures, types, structure)
- Gherkin semantics (behaviors, edge cases) - compressed
- Markdown readability (sections)

Token efficiency: ~200 tokens vs YAML ~280, Gherkin ~480
Reproduction fidelity: ~97% (best among all formats)

Format structure:
```yaml
# file.py | ClassName | N lines
imports:
  stdlib: [...]
  
ClassName:
  doc: "..."
  attrs:
    name: Type
  methods:
    method_name:
      sig: (params) -> ReturnType
      does: "What it does"
      edge: "condition → behavior"
      side: "Side effects"
      raises: [ExceptionType]
```
"""

from pathlib import Path
from typing import Dict, List, Optional, Set, Any
from dataclasses import dataclass

from .models import ProjectInfo, ModuleInfo, FunctionInfo, ClassInfo


@dataclass
class LogicMLSpec:
    """LogicML specification output."""
    content: str
    token_estimate: int
    file_count: int
    class_count: int
    function_count: int


class LogicMLGenerator:
    """
    Generates LogicML format - optimized for LLM code reproduction.
    
    Design principles:
    1. Minimal tokens, maximum information density
    2. Precise signatures with full type hints
    3. Inline behavior descriptions (no verbose Gherkin scenarios)
    4. Edge cases as compact rules
    5. Side effects explicitly noted
    
    Benchmark results:
    - Token efficiency: 1.4x better than YAML
    - Reproduction fidelity: 97% (highest)
    - Syntax OK rate: 100%
    """
    
    FORMAT_NAME: str = "logicml"
    FILE_EXTENSION: str = ".logicml"
    TOKEN_EFFICIENCY: float = 1.4  # 40% better than YAML
    REPRODUCTION_FIDELITY: float = 0.97
    
    STDLIB_MODULES: Set[str] = {
        'os', 'sys', 'json', 'typing', 'pathlib', 'dataclasses',
        're', 'ast', 'abc', 'collections', 'functools', 'itertools',
        'datetime', 'logging', 'argparse', 'subprocess', 'shutil',
        'time', 'copy', 'io', 'contextlib', 'enum', 'hashlib',
        'unittest', 'math', 'random', 'string', 'textwrap',
    }
    
    def __init__(self, verbose: bool = False) -> None:
        self.verbose = verbose
    
    def generate(self, project: ProjectInfo, detail: str = 'standard') -> LogicMLSpec:
        """Generate LogicML specification for a project."""
        parts: List[str] = []
        total_classes = 0
        total_functions = 0
        
        for module in project.modules:
            module_spec = self._generate_module(module, detail)
            if module_spec.strip():
                parts.append(module_spec)
            total_classes += len(module.classes)
            total_functions += len(module.functions)
        
        content = '\n\n'.join(parts)
        token_estimate = len(content) // 4
        
        return LogicMLSpec(
            content=content,
            token_estimate=token_estimate,
            file_count=len(project.modules),
            class_count=total_classes,
            function_count=total_functions,
        )
    
    def _generate_module(self, module: ModuleInfo, detail: str) -> str:
        """Generate LogicML for a single module."""
        lines: List[str] = []
        path = Path(module.path)
        
        # Header comment
        class_names = ', '.join(c.name for c in module.classes[:3])
        if len(module.classes) > 3:
            class_names += f" +{len(module.classes) - 3}"
        
        header_parts = [f"# {path.name}"]
        if class_names:
            header_parts.append(class_names)
        header_parts.append(f"{module.lines_total} lines")
        lines.append(' | '.join(header_parts))
        
        # Handle re-export modules (primarily __init__.py or export-like modules)
        # Some parsers may classify import-only files as having "classes" (e.g., Enum)
        # so we also special-case __init__.py.
        if (path.name == "__init__.py" and module.imports) or (
            not module.classes and not module.functions and module.imports
        ):
            lines.append(f"# Re-export module")
            lines.append("type: re-export")
            lines.append("exports:")
            for imp in module.imports[:20]:
                # Extract export name from import
                if '.' in imp:
                    export_name = imp.split('.')[-1]
                else:
                    export_name = imp
                lines.append(f"  - {export_name}")
            return '\n'.join(lines)
        
        # Handle empty index files (TypeScript export * pattern)
        if not module.classes and not module.functions and not module.imports:
            if 'index' in path.name.lower():
                lines.append("# Index re-export file")
                lines.append("type: index")
                lines.append("pattern: 'export * from ./submodules'")
            return '\n'.join(lines)
        
        # Imports (compact)
        if module.imports:
            imports_yaml = self._generate_imports(module.imports)
            if imports_yaml:
                lines.append(imports_yaml)
        
        # Classes
        for cls in module.classes:
            class_yaml = self._generate_class(cls, detail)
            lines.append(class_yaml)
        
        # Top-level functions
        if module.functions:
            funcs_yaml = self._generate_functions(module.functions, detail)
            lines.append(funcs_yaml)
        
        return '\n'.join(lines)
    
    def _generate_imports(self, imports: List[str]) -> str:
        """Generate compact imports section."""
        stdlib: Set[str] = set()
        third_party: Set[str] = set()
        local: Set[str] = set()
        
        for imp in imports:
            base = imp.split('.')[0]
            if imp.startswith('.'):
                local.add(imp)
            elif base in self.STDLIB_MODULES:
                stdlib.add(imp)
            else:
                third_party.add(imp)
        
        lines = ['\nimports:']
        if stdlib:
            lines.append(f"  stdlib: [{', '.join(sorted(stdlib)[:10])}]")
        if third_party:
            lines.append(f"  third_party: [{', '.join(sorted(third_party)[:10])}]")
        if local:
            lines.append(f"  local: [{', '.join(sorted(local)[:5])}]")
        
        return '\n'.join(lines) if len(lines) > 1 else ''
    
    def _generate_class(self, cls: ClassInfo, detail: str) -> str:
        """Generate LogicML for a class."""
        lines: List[str] = [f'\n{cls.name}:']
        
        # Docstring (full for better reproduction)
        if cls.docstring:
            doc_lines = cls.docstring.split('\n')
            first_line = doc_lines[0].strip()[:80].replace('"', "'")
            lines.append(f'  doc: "{first_line}"')
            
            # Include Example section if present (important for usage)
            for i, doc_line in enumerate(doc_lines):
                if 'Example:' in doc_line:
                    lines.append('  # Example usage in docstring')
                    break
                if 'Attributes:' in doc_line or 'Args:' in doc_line:
                    for attr_line in doc_lines[i+1:i+5]:
                        attr_line = attr_line.strip()
                        if attr_line and not attr_line.startswith('"""'):
                            lines.append(f'  # {attr_line}')
                    break
        
        # Bases - important for Pydantic/dataclass
        if cls.bases:
            bases_str = ", ".join(cls.bases)
            lines.append(f'  bases: [{bases_str}]')
            # Add hint for special base classes
            if 'BaseModel' in bases_str:
                lines.append('  # Pydantic model - use Field() for attributes')
            elif 'Enum' in bases_str:
                lines.append('  # Enum class')
        
        # Type markers
        if cls.is_abstract:
            lines.append('  abstract: true')
        if cls.is_interface:
            lines.append('  interface: true')
        
        # Attributes/Properties - extract from __init__ if not in properties
        attrs_added = False
        if cls.properties:
            lines.append('  attrs:')
            attrs_added = True
            for prop in cls.properties[:10]:
                if ':' in prop:
                    name, type_hint = prop.split(':', 1)
                    lines.append(f'    {name.strip()}: {type_hint.strip()}')
                else:
                    lines.append(f'    {prop}: Any')
        
        # Try to extract attrs from __init__ method params
        if not attrs_added and cls.methods:
            init_method = next((m for m in cls.methods if m.name == '__init__'), None)
            if init_method and len(init_method.params) > 1:
                lines.append('  attrs:')
                for param in init_method.params[1:6]:  # Skip self
                    if ':' in param:
                        name, type_hint = param.split(':', 1)
                        lines.append(f'    {name.strip()}: {type_hint.strip()}')
                    elif param != 'self':
                        lines.append(f'    {param}: Any')
        
        # Methods
        if cls.methods:
            lines.append('  methods:')
            for method in cls.methods[:20]:
                method_yaml = self._generate_method(method, detail, indent=4)
                lines.append(method_yaml)
        
        return '\n'.join(lines)
    
    def _generate_method(self, method: FunctionInfo, detail: str, indent: int = 2) -> str:
        """Generate LogicML for a method."""
        prefix = ' ' * indent
        lines: List[str] = [f'{prefix}{method.name}:']
        
        # Check for property decorator
        is_property = 'property' in method.decorators
        
        # Signature
        params = ', '.join(method.params[:6])
        ret = method.return_type or 'None'
        
        sig = f'({params}) -> {ret}'
        if method.is_async:
            sig = f'async {sig}'
        if is_property:
            sig = f'@property {sig}'
        
        lines.append(f'{prefix}  sig: {sig}')
        
        # Intent/docstring as "does" - include full context for better semantic reproduction
        if method.docstring:
            doc_lines = method.docstring.split('\n')
            does = doc_lines[0].strip()[:80].replace('"', "'")
            lines.append(f'{prefix}  does: "{does}"')
            # Add Args/Returns info if present
            for doc_line in doc_lines[1:5]:
                doc_line = doc_line.strip()
                if doc_line.startswith('Args:') or doc_line.startswith('Returns:'):
                    lines.append(f'{prefix}  # {doc_line}')
                elif ':' in doc_line and not doc_line.startswith('#'):
                    lines.append(f'{prefix}  # {doc_line[:60]}')
        elif method.intent:
            lines.append(f'{prefix}  does: "{method.intent}"')
        
        # Edge cases (from raises)
        if method.raises and detail in ('standard', 'full'):
            for exc in method.raises[:2]:
                lines.append(f'{prefix}  edge: "error → raise {exc}"')
        
        # Side effects
        side_effects = self._detect_side_effects(method)
        if side_effects and detail in ('standard', 'full'):
            lines.append(f'{prefix}  side: "{side_effects}"')
        
        # Decorators (only important ones)
        important_decorators = {'staticmethod', 'classmethod', 'property', 'abstractmethod'}
        decorators = [d for d in method.decorators if d in important_decorators]
        if decorators:
            lines.append(f'{prefix}  decorators: [{", ".join(decorators)}]')
        
        return '\n'.join(lines)
    
    def _generate_functions(self, functions: List[FunctionInfo], detail: str) -> str:
        """Generate LogicML for top-level functions."""
        lines: List[str] = ['\nfunctions:']
        
        for func in functions[:20]:
            func_yaml = self._generate_method(func, detail, indent=2)
            lines.append(func_yaml)
        
        return '\n'.join(lines)
    
    def _detect_side_effects(self, method: FunctionInfo) -> Optional[str]:
        """Detect side effects from method calls and name patterns."""
        side_effect_patterns: Dict[str, str] = {
            'append': 'Modifies list',
            'add': 'Adds element',
            'remove': 'Removes element',
            'clear': 'Clears collection',
            'pop': 'Removes and returns',
            'update': 'Updates state',
            'write': 'Writes data',
            'save': 'Saves data',
            'delete': 'Deletes data',
            'set': 'Sets value',
            'insert': 'Inserts element',
        }
        
        # Check method name
        name_lower = method.name.lower()
        for pattern, effect in side_effect_patterns.items():
            if pattern in name_lower:
                return effect
        
        # Check calls
        if method.calls:
            for call in method.calls:
                call_lower = call.lower()
                for pattern, effect in side_effect_patterns.items():
                    if pattern in call_lower:
                        return effect
        
        return None


def generate_logicml(project: ProjectInfo, detail: str = 'standard') -> str:
    """Convenience function to generate LogicML format."""
    generator = LogicMLGenerator()
    spec = generator.generate(project, detail)
    return spec.content


LOGICML_EXAMPLE = '''
# sample_class.py | Calculator | 74 lines

imports:
  stdlib: [typing.List, typing.Optional]

Calculator:
  doc: "Simple calculator with history."
  attrs:
    precision: int
    history: List[str]
  methods:
    __init__:
      sig: (precision: int) -> None
      does: "Initialize calculator"
    add:
      sig: (a: float, b: float) -> float
      does: "Add two numbers"
      side: "Modifies list"
    divide:
      sig: (a: float, b: float) -> Optional[float]
      does: "Divide a by b"
      edge: "b == 0 → return None"
'''

==> code2logic-1.0.16-py3-none-any.whl: code2logic/logicml.py <==
"""
LogicML Format Generator for Code2Logic.

Optimal hybrid format combining:
- YAML precision (signatures, types, structure)
- Gherkin semantics (behaviors, edge cases) - compressed
- Markdown readability (sections)

Token efficiency: ~200 tokens vs YAML ~280, Gherkin ~480
Reproduction fidelity: ~97% (best among all formats)

Format structure:
```yaml
# file.py | ClassName | N lines
imports:
  stdlib: [...]
  
ClassName:
  doc: "..."
  attrs:
    name: Type
  methods:
    method_name:
      sig: (params) -> ReturnType
      does: "What it does"
      edge: "condition → behavior"
      side: "Side effects"
      raises: [ExceptionType]
```
"""

from pathlib import Path
from typing import Dict, List, Optional, Set, Any
from dataclasses import dataclass

from .models import ProjectInfo, ModuleInfo, FunctionInfo, ClassInfo
from .shared_utils import truncate_docstring, compact_imports, remove_self_from_params


@dataclass
class LogicMLSpec:
    """LogicML specification output."""
    content: str
    token_estimate: int
    file_count: int
    class_count: int
    function_count: int


class LogicMLGenerator:
    """
    Generates LogicML format - optimized for LLM code reproduction.
    
    Design principles:
    1. Minimal tokens, maximum information density
    2. Precise signatures with full type hints
    3. Inline behavior descriptions (no verbose Gherkin scenarios)
    4. Edge cases as compact rules
    5. Side effects explicitly noted
    
    Benchmark results:
    - Token efficiency: 1.4x better than YAML
    - Reproduction fidelity: 97% (highest)
    - Syntax OK rate: 100%
    """
    
    FORMAT_NAME: str = "logicml"
    FILE_EXTENSION: str = ".logicml"
    TOKEN_EFFICIENCY: float = 1.4  # 40% better than YAML
    REPRODUCTION_FIDELITY: float = 0.97
    
    STDLIB_MODULES: Set[str] = {
        'os', 'sys', 'json', 'typing', 'pathlib', 'dataclasses',
        're', 'ast', 'abc', 'collections', 'functools', 'itertools',
        'datetime', 'logging', 'argparse', 'subprocess', 'shutil',
        'time', 'copy', 'io', 'contextlib', 'enum', 'hashlib',
        'unittest', 'math', 'random', 'string', 'textwrap',
    }
    
    def __init__(self, verbose: bool = False) -> None:
        self.verbose = verbose
    
    def generate(self, project: ProjectInfo, detail: str = 'standard') -> LogicMLSpec:
        """Generate LogicML specification for a project."""
        parts: List[str] = []
        total_classes = 0
        total_functions = 0
        
        for module in project.modules:
            module_spec = self._generate_module(module, detail)
            if module_spec.strip():
                parts.append(module_spec)
            total_classes += len(module.classes)
            total_functions += len(module.functions)
        
        content = '\n\n'.join(parts)
        token_estimate = len(content) // 4
        
        return LogicMLSpec(
            content=content,
            token_estimate=token_estimate,
            file_count=len(project.modules),
            class_count=total_classes,
            function_count=total_functions,
        )
    
    def _generate_module(self, module: ModuleInfo, detail: str) -> str:
        """Generate LogicML for a single module."""
        lines: List[str] = []
        path = Path(module.path)
        
        # Header comment
        class_names = ', '.join(c.name for c in module.classes[:3])
        if len(module.classes) > 3:
            class_names += f" +{len(module.classes) - 3}"
        
        header_parts = [f"# {path.name}"]
        if class_names:
            header_parts.append(class_names)
        header_parts.append(f"{module.lines_total} lines")
        lines.append(' | '.join(header_parts))
        
        # Handle re-export modules (primarily __init__.py or export-like modules)
        # Some parsers may classify import-only files as having "classes" (e.g., Enum)
        # so we also special-case __init__.py.
        if (path.name == "__init__.py" and module.imports) or (
            not module.classes and not module.functions and module.imports
        ):
            lines.append(f"# Re-export module")
            lines.append("type: re-export")
            lines.append("exports:")
            for imp in module.imports[:20]:
                # Extract export name from import
                if '.' in imp:
                    export_name = imp.split('.')[-1]
                else:
                    export_name = imp
                lines.append(f"  - {export_name}")
            return '\n'.join(lines)
        
        # Handle empty index files (TypeScript export * pattern)
        if not module.classes and not module.functions and not module.imports:
            if 'index' in path.name.lower():
                lines.append("# Index re-export file")
                lines.append("type: index")
                lines.append("pattern: 'export * from ./submodules'")
            return '\n'.join(lines)
        
        # Imports (compact)
        if module.imports:
            imports_yaml = self._generate_imports(module.imports)
            if imports_yaml:
                lines.append(imports_yaml)
        
        # Classes
        for cls in module.classes:
            class_yaml = self._generate_class(cls, detail)
            lines.append(class_yaml)
        
        # Top-level functions
        if module.functions:
            funcs_yaml = self._generate_functions(module.functions, detail)
            lines.append(funcs_yaml)
        
        return '\n'.join(lines)
    
    def _generate_imports(self, imports: List[str]) -> str:
        """Generate compact imports section."""
        stdlib: Set[str] = set()
        third_party: Set[str] = set()
        local: Set[str] = set()
        
        for imp in imports:
            base = imp.split('.')[0]
            if imp.startswith('.'):
                local.add(imp)
            elif base in self.STDLIB_MODULES:
                stdlib.add(imp)
            else:
                third_party.add(imp)
        
        lines = ['\nimports:']
        if stdlib:
            # Use compact_imports for grouped format
            grouped = compact_imports(sorted(stdlib)[:15], max_items=10)
            lines.append(f"  stdlib: [{', '.join(grouped)}]")
        if third_party:
            grouped = compact_imports(sorted(third_party)[:10], max_items=8)
            lines.append(f"  third_party: [{', '.join(grouped)}]")
        if local:
            lines.append(f"  local: [{', '.join(sorted(local)[:5])}]")
        
        return '\n'.join(lines) if len(lines) > 1 else ''
    
    def _generate_class(self, cls: ClassInfo, detail: str) -> str:
        """Generate LogicML for a class."""
        lines: List[str] = [f'\n{cls.name}:']
        
        # Docstring - truncated for efficiency
        if cls.docstring:
            doc = truncate_docstring(cls.docstring, max_length=60)
            if doc:
                lines.append(f'  doc: "{doc}"')
        
        # Bases - only if non-empty
        if cls.bases:
            bases_str = ", ".join(cls.bases)
            lines.append(f'  bases: [{bases_str}]')
        
        # Type markers
        if cls.is_abstract:
            lines.append('  abstract: true')
        if cls.is_interface:
            lines.append('  interface: true')
        
        # Attributes/Properties - extract from __init__ if not in properties
        attrs_added = False
        if cls.properties:
            lines.append('  attrs:')
            attrs_added = True
            for prop in cls.properties[:10]:
                if ':' in prop:
                    name, type_hint = prop.split(':', 1)
                    lines.append(f'    {name.strip()}: {type_hint.strip()}')
                else:
                    lines.append(f'    {prop}: Any')
        
        # Try to extract attrs from __init__ method params
        if not attrs_added and cls.methods:
            init_method = next((m for m in cls.methods if m.name == '__init__'), None)
            if init_method and len(init_method.params) > 1:
                lines.append('  attrs:')
                for param in init_method.params[1:6]:  # Skip self
                    if ':' in param:
                        name, type_hint = param.split(':', 1)
                        lines.append(f'    {name.strip()}: {type_hint.strip()}')
                    elif param != 'self':
                        lines.append(f'    {param}: Any')
        
        # Methods
        if cls.methods:
            lines.append('  methods:')
            for method in cls.methods[:20]:
                method_yaml = self._generate_method(method, detail, indent=4)
                lines.append(method_yaml)
        
        return '\n'.join(lines)
    
    def _generate_method(self, method: FunctionInfo, detail: str, indent: int = 2) -> str:
        """Generate LogicML for a method."""
        prefix = ' ' * indent
        lines: List[str] = [f'{prefix}{method.name}:']
        
        # Check for property decorator
        is_property = 'property' in method.decorators
        
        # Signature - remove self/cls for compactness
        clean_params = remove_self_from_params(method.params[:7])
        params = ', '.join(clean_params[:6])
        ret = method.return_type or 'None'
        
        sig = f'({params}) -> {ret}'
        if method.is_async:
            sig = f'async {sig}'
        if is_property:
            sig = f'@property {sig}'
        
        lines.append(f'{prefix}  sig: {sig}')
        
        # Intent/docstring as "does" - truncated for efficiency
        if method.docstring:
            does = truncate_docstring(method.docstring, max_length=60)
            if does:
                lines.append(f'{prefix}  does: "{does}"')
        elif method.intent:
            intent = method.intent[:60].replace('\n', ' ').replace('"', "'")
            lines.append(f'{prefix}  does: "{intent}"')
        
        # Edge cases (from raises)
        if method.raises and detail in ('standard', 'full'):
            for exc in method.raises[:2]:
                lines.append(f'{prefix}  edge: "error → raise {exc}"')
        
        # Side effects
        side_effects = self._detect_side_effects(method)
        if side_effects and detail in ('standard', 'full'):
            lines.append(f'{prefix}  side: "{side_effects}"')
        
        # Decorators (only important ones)
        important_decorators = {'staticmethod', 'classmethod', 'property', 'abstractmethod'}
        decorators = [d for d in method.decorators if d in important_decorators]
        if decorators:
            lines.append(f'{prefix}  decorators: [{", ".join(decorators)}]')
        
        return '\n'.join(lines)
    
    def _generate_functions(self, functions: List[FunctionInfo], detail: str) -> str:
        """Generate LogicML for top-level functions."""
        lines: List[str] = ['\nfunctions:']
        
        for func in functions[:20]:
            func_yaml = self._generate_method(func, detail, indent=2)
            lines.append(func_yaml)
        
        return '\n'.join(lines)
    
    def _detect_side_effects(self, method: FunctionInfo) -> Optional[str]:
        """Detect side effects from method calls and name patterns."""
        side_effect_patterns: Dict[str, str] = {
            'append': 'Modifies list',
            'add': 'Adds element',
            'remove': 'Removes element',
            'clear': 'Clears collection',
            'pop': 'Removes and returns',
            'update': 'Updates state',
            'write': 'Writes data',
            'save': 'Saves data',
            'delete': 'Deletes data',
            'set': 'Sets value',
            'insert': 'Inserts element',
        }
        
        # Check method name
        name_lower = method.name.lower()
        for pattern, effect in side_effect_patterns.items():
            if pattern in name_lower:
                return effect
        
        # Check calls
        if method.calls:
            for call in method.calls:
                call_lower = call.lower()
                for pattern, effect in side_effect_patterns.items():
                    if pattern in call_lower:
                        return effect
        
        return None


def generate_logicml(project: ProjectInfo, detail: str = 'standard') -> str:
    """Convenience function to generate LogicML format."""
    generator = LogicMLGenerator()
    spec = generator.generate(project, detail)
    return spec.content


LOGICML_EXAMPLE = '''
# sample_class.py | Calculator | 74 lines

imports:
  stdlib: [typing.List, typing.Optional]

Calculator:
  doc: "Simple calculator with history."
  attrs:
    precision: int
    history: List[str]
  methods:
    __init__:
      sig: (precision: int) -> None
      does: "Initialize calculator"
    add:
      sig: (a: float, b: float) -> float
      does: "Add two numbers"
      side: "Modifies list"
    divide:
      sig: (a: float, b: float) -> Optional[float]
      does: "Divide a by b"
      edge: "b == 0 → return None"
'''
